[PATCH, libfortran]: Committed: Read rounding mode from SSE mxcsr register on x86_64.

Uros Bizjak ubizjak@gmail.com
Tue Jul 23 09:56:00 GMT 2013


Hello!

On x86_64, we can look into SSE mxcsr register to determine rounding mode.


2013-07-23  Uros Bizjak  <ubizjak@gmail.com>

    * config/fpu-387.h (get_fpu_rounding_mode): Read rounding mode
    from SSE mxcsr register on x86_64.

Tested on x86_64-pc-linux-gnu {,-m32}  and committed to mainline svn.

BTW. gfortran.dg/round_4.f90 test will fail on glibc < 2.17 due to
glibc bug 3479 [1], "Incorrect rounding in strtod()".

[1] http://sourceware.org/bugzilla/show_bug.cgi?id=3479

Uros.
-------------- next part --------------
Index: config/fpu-387.h
===================================================================
--- config/fpu-387.h	(revision 201156)
+++ config/fpu-387.h	(working copy)
@@ -102,11 +102,11 @@ has_sse (void)
 /* i387 rounding modes.  */
 
 #define _FPU_RC_NEAREST 0x0
-#define _FPU_RC_DOWN    0x400
-#define _FPU_RC_UP      0x800
-#define _FPU_RC_ZERO    0xc00
+#define _FPU_RC_DOWN    0x1
+#define _FPU_RC_UP      0x2
+#define _FPU_RC_ZERO    0x3
 
-#define _FPU_RC_MASK    0xc00
+#define _FPU_RC_MASK    0x3
 
 
 void
@@ -202,8 +202,9 @@ set_fpu_rounding_mode (int round)
 
   __asm__ __volatile__ ("fnstcw\t%0" : "=m" (cw));
 
-  cw &= ~_FPU_RC_MASK;
-  cw |= round_mode;
+  /* The x87 round control bits are shifted by 10 bits.  */
+  cw &= ~(_FPU_RC_MASK << 10);
+  cw |= round_mode << 10;
 
   __asm__ __volatile__ ("fldcw\t%0" : : "m" (cw));
 
@@ -213,9 +214,9 @@ set_fpu_rounding_mode (int round)
 
       __asm__ __volatile__ ("%vstmxcsr\t%0" : "=m" (cw_sse));
 
-      /* The SSE round control bits are shifted by 3 bits.  */
-      cw_sse &= ~(_FPU_RC_MASK << 3);
-      cw_sse |= round_mode << 3;
+      /* The SSE round control bits are shifted by 13 bits.  */
+      cw_sse &= ~(_FPU_RC_MASK << 13);
+      cw_sse |= round_mode << 13;
 
       __asm__ __volatile__ ("%vldmxcsr\t%0" : : "m" (cw_sse));
     }
@@ -224,13 +225,27 @@ set_fpu_rounding_mode (int round)
 int
 get_fpu_rounding_mode (void)
 {
+  int round_mode;
+
+#ifdef __x86_64__
+  unsigned int cw;
+
+  __asm__ __volatile__ ("%vstmxcsr\t%0" : "=m" (cw));
+
+  /* The SSE round control bits are shifted by 13 bits.  */
+  round_mode = cw >> 13;
+#else
   unsigned short cw;
 
   __asm__ __volatile__ ("fnstcw\t%0" : "=m" (cw));
 
-  cw &= _FPU_RC_MASK;
+  /* The x87 round control bits are shifted by 10 bits.  */
+  round_mode = cw >> 10;
+#endif
 
-  switch (cw)
+  round_mode &= _FPU_RC_MASK;
+
+  switch (round_mode)
     {
     case _FPU_RC_NEAREST:
       return GFC_FPE_TONEAREST;


More information about the Gcc-patches mailing list